/* ===============================================================================================
 * // Ray Tracing in 2D
 * // Autor		:	faceH0r
 * // Date		:	28.04.06
 * ===============================================================================================
 */

#include		<windows.h>

#include		<stdio.h>
#include		<math.h>

#include		<stdio.h>

#include		"RayCasting.h"

#define			RC_CLASS	"Ray casting class"
#define			RC_WINDOW	"Ray casting"

// 
#define			RC_INIT		0x0001
#define			RC_UPDATE	0x0002

class			rcastMain
{

public:

	HDC		hDc;
	HWND	hWnd;

	float	fCurTime;
	float	fDelTime;
	float	fOldTime;

	float	fSpeedWalk;
	float	fSpeedRot;

	int		iState;

	char	chKeys[ 256 ];

	clRayCast	rc;

	void	funSetState	( int iState );
	void	funUpdate	( void );

private:

};

void rcastMain::funSetState( int iState )
{
	this->iState = iState;
}

void rcastMain::funUpdate( void )
{
	switch( iState )
	{
	case RC_INIT:
		hDc = GetDC( hWnd );
		rc.rcInit( hDc, 320, 240 );
		rc.rcInitDir();
		rc.rcClearColor4b( 100, 0, 0, 255 );
		rc.rcSetCount( 32 );

		fSpeedWalk = 0.003f;
		fSpeedRot  = 0.003f;

		rc.vEye.v[ 1 ] = - 4;

		iState = RC_UPDATE;

		break;
	case RC_UPDATE:
		rc.rcClear( RC_COLOR_BUFFER_BIT );

		if( chKeys[ VK_ESCAPE ] )
		{
			PostQuitMessage( 0 );
		}

		if( chKeys[ VK_LEFT ] )
		{
			rc.rcCamRot( fSpeedRot * fDelTime );
		}

		if( chKeys[ VK_RIGHT ] )
		{
			rc.rcCamRot( - fSpeedRot * fDelTime );
		}

		if( chKeys[ VK_UP ] )
		{
			rc.vEye.v[ 0 ] += rc.vDir.v[ 0 ] * fSpeedWalk * fDelTime;
			rc.vEye.v[ 1 ] += rc.vDir.v[ 1 ] * fSpeedWalk * fDelTime;
		}

		if( chKeys[ VK_DOWN ] )
		{
			rc.vEye.v[ 0 ] -= rc.vDir.v[ 0 ] * fSpeedWalk * fDelTime;
			rc.vEye.v[ 1 ] -= rc.vDir.v[ 1 ] * fSpeedWalk * fDelTime;
		}

		rc.rcColor4b( 255, 0, 0, 255 );
		rc.rcLine( -5.0f, 5.0f, 5.0f, 5.0f );
		rc.rcColor4b( 0, 255, 0, 255 );
		rc.rcLine( 5.0f, -5.0f, -5.0f, -5.0f );
		rc.rcColor4b( 0, 0, 255, 255 );
		rc.rcLine( -5.0f, 5.0f, -5.0f, -5.0f );
		rc.rcColor4b( 255, 255, 0, 255 );
		rc.rcLine( 5.0f, -5.0f, 5.0f, 5.0f );

		rc.rcColor4b( 200, 200, 200, 255 );
		rc.rcLine( 1.0f, 1.0f, -1.0f, 1.0f );
		rc.rcColor4b( 100, 100, 100, 255 );
		rc.rcLine( -1.0f, -1.0f, 1.0f, -1.0f );
		rc.rcColor4b( 250, 250, 250, 255 );
		rc.rcLine( -1.0f, -1.0f, -1.0f, 1.0f );
		rc.rcColor4b( 150, 150, 150, 255 );
		rc.rcLine( 1.0f, 1.0f, 1.0f, -1.0f );

		rc.rcColor4b( 200, 2, 200, 255 );
		rc.rcLine( 4.0f, 1.0f, 2.0f, 0.0f );
		rc.rcColor4b( 100, 1, 100, 255 );
		rc.rcLine( 2.0f, 0.0f, 4.0f, -1.0f );
		rc.rcColor4b( 250, 2, 250, 255 );
		rc.rcLine( 2.0f, -1.0f, 2.0f, 1.0f );

		rc.rcColor4b( 50, 100, 0, 255 );
		rc.rcCircle( 4.0f, 4.0f, 0.05f );
		rc.rcColor4b( 100, 0, 100, 255 );
		rc.rcCircle( 3.0f, 3.0f, 0.15f );
		rc.rcColor4b( 150, 0, 0, 255 );
		rc.rcCircle( 2.0f, 2.0f, 0.25f );

		rc.rcColor4b( 120, 120, 120, 255 );
		rc.rcCircle( -4.0f, 1.0f, 0.5f );
		rc.rcColor4b( 170, 170, 170, 255 );
		rc.rcCircle( -3.0f, 1.0f, 0.1f );

		rc.rcTrace();
		rc.rcSwapBuffers();

		break;
	}
}

rcastMain	rcast;

LRESULT CALLBACK WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

LRESULT CALLBACK WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	case WM_KEYDOWN:
		rcast.chKeys[ wParam ] = true;

		break;
	case WM_KEYUP:
		rcast.chKeys[ wParam ] = false;

		break;
	default:
		return DefWindowProc( hWnd, uMsg, wParam, lParam );

		break;
	}
}

WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowSmd )
{
	WNDCLASSEX		wClass;
	MSG				msg;

	memset( &wClass, 0, sizeof( wClass ) );
	wClass.cbSize = sizeof( wClass );
	wClass.lpszClassName = RC_CLASS;
	wClass.lpfnWndProc   = WndProc;

	RegisterClassEx( &wClass );

	rcast.hWnd = CreateWindowEx( 0, RC_CLASS, RC_WINDOW, WS_POPUP, 0, 0, 320, 240, 0, 0, 0, 0 );

	ShowWindow( rcast.hWnd, SW_SHOW );

	rcast.funSetState( RC_INIT );

	ShowCursor( false );

	rcast.fOldTime = GetTickCount();

	DEVMODE		dm;

	memset( &dm, 0, sizeof( dm ) );

	dm.dmSize = sizeof( dm );

	dm.dmPelsWidth  = 320;
	dm.dmPelsHeight = 240;
	dm.dmBitsPerPel = 32;
	
	dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	ChangeDisplaySettings( &dm, CDS_FULLSCREEN );

	while( msg.message != WM_QUIT)
	{
		if( PeekMessage( &msg, 0, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}

		rcast.fCurTime = GetTickCount();
		rcast.fDelTime = fabs( rcast.fCurTime - rcast.fOldTime );
		rcast.fOldTime = rcast.fCurTime;

		rcast.funUpdate();
	}

	return 0;
}