#include		"RayCasting.h"

clRayCast::clRayCast( void )
{
}

clRayCast::~clRayCast( void )
{
	free( shpData );
	free( bufData );
	free( dirData );
}

void clRayCast::rcInit( HDC hDc, RCINT iSizeX, RCINT iSizeY )
{
	rlist.iSSizeX = iSizeX;
	rlist.iSSizeY = iSizeY;
	rlist.iSSizeM = iSizeX * iSizeY;
	rlist.iSSizeM4 = ( rlist.iSSizeM << 2 );
	rlist.iSSizeXDiv2 = ( iSizeX >> 1 );
	rlist.iSSizeYDiv2 = ( iSizeY >> 1 );

	rlist.hDc = hDc;

	bufData = new char[ rlist.iSSizeM4 ];
	dirData = new strcRCVector2f[ rlist.iSSizeX ];
	rlist.hBMP   = CreateBitmap( rlist.iSSizeX, rlist.iSSizeY, 1, 32, bufData );
	rlist.hDcBMP = CreateCompatibleDC( rlist.hDc );
	SelectObject( rlist.hDcBMP, rlist.hBMP );
}

void clRayCast::rcInitDir( void )
{
	int		x;

	vDir.v[ 0 ] = 0.0f;
	vDir.v[ 1 ] = 1.0f;

	for( x = 0; x < rlist.iSSizeX; x ++ )
	{
		dirData[ x ].v[ 0 ]	= x - rlist.iSSizeXDiv2;
		dirData[ x ].v[ 1 ]	= 250;

		rcVecNormalize( dirData[ x ] );
	}
}

void clRayCast::rcClear( RCINT rtiMode )
{
	if( rtiMode && RC_COLOR_BUFFER_BIT )
	{
		for( long l = 0; l < rlist.iSSizeM; l ++ )
		{
			bufDataUInt[ l ] = ( int & )colBack.c;
		}
	}
}

void clRayCast::rcClearColor4b(  RCUCHAR r, RCUCHAR g, RCUCHAR b, RCUCHAR a )
{
	colBack.c[ 0 ] = b;
	colBack.c[ 1 ] = g;
	colBack.c[ 2 ] = r;
	colBack.c[ 3 ] = a;
}

void clRayCast::rcColor4b( RCUCHAR r, RCUCHAR g, RCUCHAR b, RCUCHAR a )
{
	colCurrent.c[ 0 ] = b;
	colCurrent.c[ 1 ] = g;
	colCurrent.c[ 2 ] = r;
	colCurrent.c[ 3 ] = a;
}

void clRayCast::rcSetCount( RCINT iCountShapes )
{
	shpData = new strcRCShape[ iCountShapes ];
	this->iCountShapes = iCountShapes;
}

void clRayCast::rcLine( RCFLOAT x0, RCFLOAT y0, RCFLOAT x1, RCFLOAT y1 )
{
	if( iActiveShape + 1 < iCountShapes )
	{
		x0 += RC_DEPTH_CORRECT;
		y0 += RC_DEPTH_CORRECT;

		shpData[ iActiveShape ].lin.vv[ 0 ].v[ 0 ] = x0;
		shpData[ iActiveShape ].lin.vv[ 0 ].v[ 1 ] = y0;
		shpData[ iActiveShape ].lin.vv[ 1 ].v[ 0 ] = x1;
		shpData[ iActiveShape ].lin.vv[ 1 ].v[ 1 ] = y1;

		shpData[ iActiveShape ].lin.vc.v[ 0 ] = ( x0 + x1 ) / 2.0f;
		shpData[ iActiveShape ].lin.vc.v[ 1 ] = ( y0 + y1 ) / 2.0f;

		shpData[ iActiveShape ].lin.vn.v[ 0 ] = ( y1 - y0 );
		shpData[ iActiveShape ].lin.vn.v[ 1 ] = ( x1 - x0 );

		rcVecNormalize( shpData[ iActiveShape ].lin.vn );

		shpData[ iActiveShape ].cc = colCurrent;

		shpData[ iActiveShape ].iType = RC_LINE_SEGMENT;

		iActiveShape ++;
	}
}

void clRayCast::rcCircle( RCFLOAT x, RCFLOAT y, RCFLOAT r )
{
	if( iActiveShape + 1 < iCountShapes )
	{
		shpData[ iActiveShape ].crc.vv.v[ 0 ] = x;
		shpData[ iActiveShape ].crc.vv.v[ 1 ] = y;
		shpData[ iActiveShape ].crc.fRad      = r;

		shpData[ iActiveShape ].cc = colCurrent;

		shpData[ iActiveShape ].iType = RC_CIRCLE;

		iActiveShape ++;
	}
}

void clRayCast::rcTrace( void )
{
	int		x;
	int		y;
	int		hy;

	for( x = 0; x < rlist.iSSizeX; x ++ )
	{
		int		iIntersect = - 1;
		float	fDotProd;
		int		iMinX, iMaxX, iMinY, iMaxY;
		float	fLen;
		strcRCVector2f	vf;

		float	fSaveDepth = 1000;

		for( y = 0; y < iActiveShape; y ++ )
		{
			if( shpData[ y ].iType == RC_LINE_SEGMENT )
			{
				fDotProd = shpData[ y ].lin.vn.v[ 0 ] * dirData[ x ].v[ 0 ] + 
					       shpData[ y ].lin.vn.v[ 1 ] * dirData[ x ].v[ 1 ];

				if( fDotProd < 0.001f )
				{
					continue;
				}

				fLen = ( - ( shpData[ y ].lin.vn.v[ 0 ] * vEye.v[ 0 ] + 
					         shpData[ y ].lin.vn.v[ 1 ] * vEye.v[ 1 ] ) +
							 shpData[ y ].lin.vn.v[ 0 ] * shpData[ y ].lin.vc.v[ 0 ] +
							 shpData[ y ].lin.vn.v[ 1 ] * shpData[ y ].lin.vc.v[ 1 ] ) / fDotProd;

				vf.v[ 0 ] = vEye.v[ 0 ] + dirData[ x ].v[ 0 ] * fLen;
				vf.v[ 1 ] = vEye.v[ 1 ] + dirData[ x ].v[ 1 ] * fLen;

				iMinX = 0;
				iMaxX = 1;

				if( shpData[ y ].lin.vv[ 0 ].v[ 0 ] > shpData[ y ].lin.vv[ 1 ].v[ 0 ] )
				{
					iMinX = 1;
					iMaxX = 0;
				}

				iMinY = 0;
				iMaxY = 1;

				if( shpData[ y ].lin.vv[ 0 ].v[ 1 ] > shpData[ y ].lin.vv[ 1 ].v[ 1 ] )
				{
					iMinY = 1;
					iMaxY = 0;
				}

				if( vf.v[ 0 ] < shpData[ y ].lin.vv[ iMinX ].v[ 0 ] || vf.v[ 0 ] > shpData[ y ].lin.vv[ iMaxX ].v[ 0 ] || 
					vf.v[ 1 ] < shpData[ y ].lin.vv[ iMinY ].v[ 1 ] || vf.v[ 1 ] > shpData[ y ].lin.vv[ iMaxY ].v[ 1 ] )
				{
					continue;
				}

				if( fLen < 0 )
				{
					continue;
				}

				if( fLen < fSaveDepth )
				{
					iIntersect = y;
					hy = rlist.iSSizeY / fLen;
					fSaveDepth = fLen;
					if( hy >= rlist.iSSizeY )
						hy = rlist.iSSizeY;
				}
			}

			if( shpData[ y ].iType == RC_CIRCLE )
			{
				vf.v[ 0 ] = shpData[ y ].crc.vv.v[ 0 ] - vEye.v[ 0 ];
				vf.v[ 1 ] = shpData[ y ].crc.vv.v[ 1 ] - vEye.v[ 1 ];

				float	fDotProd = vf.v[ 0 ] * dirData[ x ].v[ 0 ] + vf.v[ 1 ] * dirData[ x ].v[ 1 ];

				if( fDotProd < 0.0f )
				{
					continue;
				}

				fLen = shpData[ y ].crc.fRad * shpData[ y ].crc.fRad - 
					( vf.v[ 0 ] * vf.v[ 0 ] + vf.v[ 1 ] * vf.v[ 1 ] ) + fDotProd * fDotProd;

				if( fLen < 0.0f )
				{
					continue;
				}

				float f = fDotProd - sqrt( fLen );

				if( f < fSaveDepth )
				{
					iIntersect = y;
					hy = rlist.iSSizeY / f;
					fSaveDepth = f;
					if( hy >= rlist.iSSizeY )
						hy = rlist.iSSizeY;
				}
			}
		}

		if( iIntersect != - 1 )
		{
			for( y = 0; y < hy; y ++ )
			{ 
				int		iRes = ( int & )shpData[ iIntersect ].cc;
				bufDataUInt[ x + ( y + rlist.iSSizeYDiv2 - int( hy / 2 ) ) * rlist.iSSizeX ] = iRes;
			}
		}
	}

	iActiveShape = 0;
}

void clRayCast::rcSwapBuffers( void )
{
	SetBitmapBits( rlist.hBMP, rlist.iSSizeM4, bufData );
	BitBlt( rlist.hDc, 0, 0, rlist.iSSizeX, rlist.iSSizeY, rlist.hDcBMP, 0, 0, SRCCOPY );
}

void clRayCast::rcCamRot( RCFLOAT ang )
{
	int		x;
	float	c = cos( ang );
	float	s = sin( ang );
	float	xt;

	for( x = 0; x < rlist.iSSizeX; x ++ )
	{
		xt = dirData[ x ].v[ 0 ];
		dirData[ x ].v[ 0 ] = xt * c - dirData[ x ].v[ 1 ] * s;
		dirData[ x ].v[ 1 ] = xt * s + dirData[ x ].v[ 1 ] * c;
	}

	rcVecRotate( ang, vDir );
}

inline void clRayCast::rcVecNormalize( strcRCVector2f &v2f )
{
	float	f1DL = 1.0f / sqrt( v2f.v[ 0 ] * v2f.v[ 0 ] + v2f.v[ 1 ] * v2f.v[ 1 ] );

	v2f.v[ 0 ] *= f1DL;
	v2f.v[ 1 ] *= f1DL;
}

inline void clRayCast::rcVecRotate( RCFLOAT ang, strcRCVector2f &v2f )
{
	float	x = v2f.v[ 0 ];
	float	y = v2f.v[ 1 ];

	v2f.v[ 0 ] = x * cos( ang ) - y * sin( ang );
	v2f.v[ 1 ] = x * sin( ang ) + y * cos( ang );
}

inline float clRayCast::rcVecLength( strcRCVector2f v2f )
{
	return ( sqrt( v2f.v[ 0 ] * v2f.v[ 0 ] + v2f.v[ 1 ] * v2f.v[ 1 ] ) );
}

inline strcRCVector2f clRayCast::rcVecSubtract( strcRCVector2f v0f, strcRCVector2f v1f )
{
	strcRCVector2f	vR;
	
	vR.v[ 0 ] = v0f.v[ 0 ] - v1f.v[ 0 ];
	vR.v[ 1 ] = v0f.v[ 1 ] - v1f.v[ 1 ];

	return vR;
}